@@ -1,11 +1,15 @@
 use crate::{
-    de::{deserialize::ReceivedPacket, PacketReader},
+    de::{
+        deserialize::{ConnAck, ReceivedPacket},
+        PacketReader,
+    },
+    network_manager::InterfaceHolder,
     ser::serialize,
     session_state::SessionState,
-    Property, {debug, error, info},
+    Error, Property, ProtocolError, {debug, error, info},
 };
 
-use embedded_nal::{nb, IpAddr, SocketAddr, TcpClientStack};
+use embedded_nal::{IpAddr, SocketAddr, TcpClientStack};
 
 use heapless::String;
 
@@ -15,48 +19,13 @@ use core::str::FromStr;
 #[derive(Debug, Copy, Clone, PartialEq)]
 pub enum QoS {
     /// A packet will be delivered at most once, but may not be delivered at all.
-    AtMostOnce,
+    AtMostOnce = 0,
 
     /// A packet will be delivered at least one time, but possibly more than once.
-    AtLeastOnce,
+    AtLeastOnce = 1,
 
     /// A packet will be delivered exactly one time.
-    ExactlyOne,
-}
-
-/// Possible errors encountered during an MQTT connection.
-#[derive(Debug, Copy, Clone, PartialEq)]
-pub enum Error<E> {
-    Network(E),
-    WriteFail,
-    NotReady,
-    Unsupported,
-    ProvidedClientIdTooLong,
-    Failed(u8),
-    Protocol(ProtocolError),
-    SessionReset,
-}
-
-/// Errors that are specific to the MQTT protocol implementation.
-#[derive(Debug, Copy, Clone, PartialEq)]
-pub enum ProtocolError {
-    Bounds,
-    DataSize,
-    Invalid,
-    Failed,
-    PacketSize,
-    MalformedPacket,
-    MalformedInteger,
-    UnknownProperty,
-    UnsupportedPacket,
-    BufferSize,
-    InvalidProperty,
-}
-
-impl<E> From<E> for Error<E> {
-    fn from(e: E) -> Error<E> {
-        Error::Network(e)
-    }
+    ExactlyOnce = 2,
 }
 
 mod sm {
@@ -67,8 +36,10 @@ mod sm {
         transitions: {
             *Restart + GotSocket = ConnectTransport,
             ConnectTransport + Connect = ConnectBroker,
-            ConnectBroker + SentConnect = Active,
+            ConnectBroker + SentConnect = Establishing,
             ConnectBroker + Disconnect = Restart,
+            Establishing + ReceivedConnAck = Active,
+            Establishing + Disconnect = Restart,
             Active + Disconnect = Restart,
         }
     }
@@ -81,29 +52,37 @@ mod sm {
 use sm::{Context, Events, StateMachine, States};
 
 /// The general structure for managing MQTT via Minimq.
-pub struct Minimq<N, const T: usize>
+pub struct Minimq<TcpStack, Clock, const MSG_SIZE: usize, const MSG_COUNT: usize>
 where
-    N: TcpClientStack,
+    TcpStack: TcpClientStack,
+    Clock: embedded_time::Clock,
 {
-    pub client: MqttClient<N, T>,
-    packet_reader: PacketReader<T>,
+    pub client: MqttClient<TcpStack, Clock, MSG_SIZE, MSG_COUNT>,
+    packet_reader: PacketReader<MSG_SIZE>,
 }
 
 /// A client for interacting with an MQTT Broker.
-pub struct MqttClient<N: TcpClientStack, const T: usize> {
-    network_stack: N,
-    socket: Option<N::TcpSocket>,
-    session_state: SessionState,
+pub struct MqttClient<
+    TcpStack: TcpClientStack,
+    Clock: embedded_time::Clock,
+    const MSG_SIZE: usize,
+    const MSG_COUNT: usize,
+> {
+    pub(crate) network: InterfaceHolder<TcpStack>,
+    clock: Clock,
+    session_state: SessionState<Clock, MSG_SIZE, MSG_COUNT>,
     connection_state: StateMachine<Context>,
 }
 
-impl<N, const T: usize> MqttClient<N, T>
+impl<TcpStack, Clock, const MSG_SIZE: usize, const MSG_COUNT: usize>
+    MqttClient<TcpStack, Clock, MSG_SIZE, MSG_COUNT>
 where
-    N: TcpClientStack,
+    TcpStack: TcpClientStack,
+    Clock: embedded_time::Clock,
 {
-    fn process(&mut self) -> Result<(), Error<N::Error>> {
+    fn process(&mut self) -> Result<(), Error<TcpStack::Error>> {
         // Potentially update the state machine depending on the current socket connection status.
-        if !self.socket_is_connected()? {
+        if !self.network.tcp_connected()? {
             self.connection_state.process_event(Events::Disconnect).ok();
         } else {
             self.connection_state.process_event(Events::Connect).ok();
@@ -112,12 +91,8 @@ where
         match self.connection_state.state() {
             // In the RESTART state, we need to reopen the TCP socket.
             &States::Restart => {
-                if self.socket.is_some() {
-                    self.network_stack.close(self.socket.take().unwrap())?;
-                }
+                self.network.allocate_socket()?;
 
-                // Allocate a new socket to use and begin connecting it.
-                self.socket.replace(self.network_stack.socket()?);
                 self.connection_state
                     .process_event(Events::GotSocket)
                     .unwrap();
@@ -125,79 +100,73 @@ where
 
             // In the connect transport state, we need to connect our TCP socket to the broker.
             &States::ConnectTransport => {
-                self.network_stack
-                    .connect(
-                        self.socket.as_mut().unwrap(),
-                        SocketAddr::new(self.session_state.broker, 1883),
-                    )
-                    .map_err(|err| match err {
-                        nb::Error::WouldBlock => Error::WriteFail,
-                        nb::Error::Other(err) => Error::Network(err),
-                    })?;
+                self.network
+                    .connect(SocketAddr::new(self.session_state.broker, 1883))?;
             }
 
             // Next, connect to the broker via the MQTT protocol.
             &States::ConnectBroker => {
-                self.reset();
-
                 let properties = [
                     // Tell the broker our maximum packet size.
-                    Property::MaximumPacketSize(T as u32),
+                    Property::MaximumPacketSize(MSG_SIZE as u32),
                     // The session does not expire.
                     Property::SessionExpiryInterval(u32::MAX),
                 ];
 
-                let mut buffer: [u8; T] = [0; T];
+                let mut buffer: [u8; MSG_SIZE] = [0; MSG_SIZE];
                 let packet = serialize::connect_message(
                     &mut buffer,
                     self.session_state.client_id.as_str().as_bytes(),
-                    self.session_state.keep_alive_interval,
+                    self.session_state.keepalive_interval(),
                     &properties,
                     // Only perform a clean start if we do not have any session state.
                     !self.session_state.is_present(),
-                )
-                .map_err(|e| Error::Protocol(e))?;
+                )?;
 
                 info!("Sending CONNECT");
-                self.write(packet)?;
+                self.network.write(packet)?;
 
                 self.connection_state
                     .process_event(Events::SentConnect)
                     .unwrap();
             }
 
+            &States::Establishing => {}
+
             _ => {}
         }
 
+        self.handle_timers()?;
+
         Ok(())
     }
 
-    fn read(&mut self, mut buf: &mut [u8]) -> Result<usize, Error<N::Error>> {
-        // Atomically access the socket.
-        let socket = self.socket.as_mut().unwrap();
-        let result = self.network_stack.receive(socket, &mut buf);
-
-        result.or_else(|err| match err {
-            nb::Error::WouldBlock => Ok(0),
-            nb::Error::Other(err) => Err(Error::Network(err)),
-        })
+    fn reset(&mut self) {
+        self.connection_state.process_event(Events::Disconnect).ok();
     }
 
-    fn write(&mut self, buf: &[u8]) -> Result<(), Error<N::Error>> {
-        let socket = self.socket.as_mut().unwrap();
-        self.network_stack
-            .send(socket, &buf)
-            .map_err(|err| match err {
-                nb::Error::WouldBlock => Error::WriteFail,
-                nb::Error::Other(err) => Error::Network(err),
-            })
-            .and_then(|written| {
-                if written != buf.len() {
-                    Err(Error::WriteFail)
-                } else {
-                    Ok(())
-                }
-            })
+    /// Configure the MQTT keep-alive interval.
+    ///
+    /// # Note
+    /// This must be completed before connecting to a broker.
+    ///
+    /// # Note
+    /// The broker may override the requested keep-alive interval. Any value requested by the
+    /// broker will be used instead.
+    ///
+    /// # Args
+    /// * `interval` - The keep-alive interval in seconds. A ping will be transmitted if no other
+    /// messages are sent within 50% of the keep-alive interval.
+    pub fn set_keepalive_interval(
+        &mut self,
+        interval_seconds: u16,
+    ) -> Result<(), Error<TcpStack::Error>> {
+        if self.connection_state.state() != &States::Active {
+            return Err(Error::NotReady);
+        }
+
+        self.session_state.set_keepalive(interval_seconds);
+        Ok(())
     }
 
     /// Subscribe to a topic.
@@ -213,18 +182,17 @@ where
         &mut self,
         topic: &'a str,
         properties: &[Property<'b>],
-    ) -> Result<(), Error<N::Error>> {
-        if !self.session_state.connected {
+    ) -> Result<(), Error<TcpStack::Error>> {
+        if self.connection_state.state() != &States::Active {
             return Err(Error::NotReady);
         }
 
         let packet_id = self.session_state.get_packet_identifier();
 
-        let mut buffer: [u8; T] = [0; T];
-        let packet = serialize::subscribe_message(&mut buffer, topic, packet_id, properties)
-            .map_err(|e| Error::Protocol(e))?;
+        let mut buffer: [u8; MSG_SIZE] = [0; MSG_SIZE];
+        let packet = serialize::subscribe_message(&mut buffer, topic, packet_id, properties)?;
 
-        self.write(packet).and_then(|_| {
+        self.network.write(packet).and_then(|_| {
             info!("Subscribing to `{}`: {}", topic, packet_id);
             self.session_state
                 .pending_subscriptions
@@ -247,8 +215,24 @@ where
     ///
     /// # Returns
     /// True if the client is connected to the broker.
-    pub fn is_connected(&mut self) -> Result<bool, Error<N::Error>> {
-        Ok(self.socket_is_connected()? && self.session_state.connected)
+    pub fn is_connected(&mut self) -> bool {
+        self.connection_state.state() == &States::Active
+    }
+
+    /// Get the count of unacknowledged QoS 1 messages.
+    ///
+    /// # Returns
+    /// Number of pending messages with QoS 1.
+    pub fn pending_messages(&self, qos: QoS) -> usize {
+        self.session_state.pending_messages(qos)
+    }
+
+    /// Determine if the client is able to process QoS 1 publish requess.
+    ///
+    /// # Returns
+    /// True if the client is able to service QoS 1 requests.
+    pub fn can_publish(&self, qos: QoS) -> bool {
+        self.session_state.can_publish(qos)
     }
 
     /// Publish a message over MQTT.
@@ -257,7 +241,7 @@ where
     /// If the client is not yet connected to the broker, the message will be silently ignored.
     ///
     /// # Note
-    /// Currently, Only QoS level 1 (at most once) delivery is supported.
+    /// Currently, QoS level 2 (exactly once) delivery is not supported.
     ///
     /// # Args
     /// * `topic` - The topic to publish the message to.
@@ -265,18 +249,18 @@ where
     /// * `qos` - The desired quality-of-service level of the message. Must be QoS::AtMostOnce
     /// * `properties` - A list of properties to associate with the message being published. May be
     ///   empty.
-    pub fn publish<'a, 'b>(
+    pub fn publish(
         &mut self,
-        topic: &'a str,
+        topic: &str,
         data: &[u8],
         qos: QoS,
-        properties: &[Property<'a>],
-    ) -> Result<(), Error<N::Error>> {
-        // TODO: QoS support.
-        assert!(qos == QoS::AtMostOnce);
+        properties: &[Property],
+    ) -> Result<(), Error<TcpStack::Error>> {
+        // TODO: QoS 2 support.
+        assert!(qos != QoS::ExactlyOnce);
 
         // If we are not yet connected to the broker, we can't transmit a message.
-        if self.is_connected()? == false {
+        if self.is_connected() == false {
             return Ok(());
         }
 
@@ -285,81 +269,104 @@ where
             topic, data, properties
         );
 
-        let mut buffer: [u8; T] = [0; T];
-        let packet = serialize::publish_message(&mut buffer, topic, data, properties)
-            .map_err(|e| Error::Protocol(e))?;
-        self.write(packet)
-    }
+        // If QoS 0 the ID will be ignored
+        let id = self.session_state.get_packet_identifier();
 
-    fn socket_is_connected(&mut self) -> Result<bool, N::Error> {
-        if self.socket.is_none() {
-            return Ok(false);
+        let mut buffer: [u8; MSG_SIZE] = [0; MSG_SIZE];
+        let packet = serialize::publish_message(&mut buffer, topic, data, qos, id, properties)?;
+
+        self.network.write(packet)?;
+        self.session_state.increment_packet_identifier();
+
+        if qos == QoS::AtLeastOnce {
+            self.session_state.handle_publish(qos, id, packet);
         }
 
-        let socket = self.socket.as_ref().unwrap();
-        self.network_stack.is_connected(socket)
+        Ok(())
     }
 
-    fn reset(&mut self) {
-        self.session_state.connected = false;
+    fn handle_connection_acknowledge(
+        &mut self,
+        acknowledge: ConnAck,
+    ) -> Result<(), Error<TcpStack::Error>> {
+        if self.connection_state.state() != &States::Establishing {
+            return Err(Error::Protocol(ProtocolError::Invalid));
+        }
+
+        let mut result = Ok(());
+
+        if acknowledge.reason_code != 0 {
+            return Err(Error::Failed(acknowledge.reason_code));
+        }
+
+        if !acknowledge.session_present {
+            if self.session_state.is_present() {
+                result = Err(Error::SessionReset);
+            }
+
+            // Reset the session state upon connection with a broker that doesn't have a
+            // session state saved for us.
+            self.session_state.reset();
+        }
+
+        self.connection_state
+            .process_event(Events::ReceivedConnAck)
+            .unwrap();
+
+        for property in acknowledge.properties {
+            match property {
+                Property::MaximumPacketSize(size) => {
+                    self.session_state.maximum_packet_size.replace(size);
+                }
+                Property::AssignedClientIdentifier(id) => {
+                    self.session_state.client_id =
+                        String::from_str(id).or(Err(Error::ProvidedClientIdTooLong))?;
+                }
+                Property::ServerKeepAlive(keep_alive) => {
+                    self.session_state.set_keepalive(keep_alive);
+                }
+                _prop => info!("Ignoring property: {:?}", _prop),
+            };
+        }
+
+        // Now that we are connected, we have session state that will be persisted.
+        self.session_state
+            .register_connection(self.clock.try_now()?);
+
+        // Replay QoS 1 messages
+        for key in self.session_state.pending_publish_ordering.iter() {
+            let message = self.session_state.pending_publish.get(&key).unwrap();
+            self.network.write(message)?;
+        }
+
+        result
     }
 
     fn handle_packet<'a, F>(
         &mut self,
         packet: ReceivedPacket<'a>,
         f: &mut F,
-    ) -> Result<(), Error<N::Error>>
+    ) -> Result<(), Error<TcpStack::Error>>
     where
-        F: FnMut(&mut MqttClient<N, T>, &'a str, &[u8], &[Property<'a>]),
+        F: FnMut(
+            &mut MqttClient<TcpStack, Clock, MSG_SIZE, MSG_COUNT>,
+            &'a str,
+            &[u8],
+            &[Property<'a>],
+        ),
     {
-        if !self.session_state.connected {
-            if let ReceivedPacket::ConnAck(acknowledge) = packet {
-                let mut result = Ok(());
+        // ConnAck packets are received outside of the connection state.
+        if let ReceivedPacket::ConnAck(ack) = packet {
+            return self.handle_connection_acknowledge(ack);
+        }
 
-                if acknowledge.reason_code != 0 {
-                    return Err(Error::Failed(acknowledge.reason_code));
-                }
-
-                if !acknowledge.session_present {
-                    if self.session_state.connected {
-                        result = Err(Error::SessionReset);
-                    }
-
-                    // Reset the session state upon connection with a broker that doesn't have a
-                    // session state saved for us.
-                    self.session_state.reset();
-                }
-
-                // Now that we are connected, we have session state that will be persisted.
-                self.session_state.register_connection();
-                self.session_state.connected = true;
-
-                for property in acknowledge.properties {
-                    match property {
-                        Property::MaximumPacketSize(size) => {
-                            self.session_state.maximum_packet_size.replace(size);
-                        }
-                        Property::AssignedClientIdentifier(id) => {
-                            self.session_state.client_id =
-                                String::from_str(id).or(Err(Error::ProvidedClientIdTooLong))?;
-                        }
-                        Property::ServerKeepAlive(keep_alive) => {
-                            self.session_state.keep_alive_interval = keep_alive;
-                        }
-                        _prop => info!("Ignoring property: {:?}", _prop),
-                    };
-                }
-
-                return result;
-            } else {
-                // It is a protocol error to receive anything else when not connected.
-                // TODO: Verify it is a protocol error.
-                error!(
-                    "Received invalid packet outside of connected state: {:?}",
-                    packet
-                );
-                return Err(Error::Protocol(ProtocolError::Invalid));
-            }
+        // All other packets must be received in the active state.
+        if !self.is_connected() {
+            error!(
+                "Received invalid packet outside of connected state: {:?}",
+                packet
+            );
+            return Err(Error::Protocol(ProtocolError::Invalid));
         }
 
         match packet {
@@ -370,6 +377,13 @@ where
                 Ok(())
             }
 
+            ReceivedPacket::PubAck(ack) => {
+                // No matter the status code the message is considered acknowledged at this point
+                self.session_state.handle_puback(ack.packet_identifier);
+
+                Ok(())
+            }
+
             ReceivedPacket::SubAck(subscribe_acknowledge) => {
                 match self
                     .session_state
@@ -391,12 +405,56 @@ where
                 Ok(())
             }
 
+            ReceivedPacket::PingResp => {
+                // Cancel the ping response timeout.
+                self.session_state.register_ping_response();
+                Ok(())
+            }
+
             _ => Err(Error::Unsupported),
         }
     }
+
+    fn handle_timers(&mut self) -> Result<(), Error<TcpStack::Error>> {
+        // If we are not connected, there's no session state to manage.
+        if self.connection_state.state() != &States::Active {
+            return Ok(());
+        }
+
+        let now = self.clock.try_now()?;
+
+        // Note: The ping timeout is set at this point so that it's running even if we fail
+        // to write the ping message. This is intentional incase the underlying transport
+        // mechanism has stalled. The ping timeout will then allow us to recover the
+        // underlying TCP connection.
+        match self.session_state.handle_ping(now) {
+            Err(()) => {
+                self.connection_state.process_event(Events::Disconnect).ok();
+            }
+
+            Ok(true) => {
+                let mut buffer: [u8; MSG_SIZE] = [0; MSG_SIZE];
+
+                // Note: If we fail to serialize or write the packet, the ping timeout timer is
+                // still running, so we will recover the TCP connection in the future.
+                let packet = serialize::ping_req_message(&mut buffer)?;
+                self.network.write(packet)?;
+            }
+
+            Ok(false) => {}
+        }
+
+        Ok(())
+    }
 }
 
-impl<N: TcpClientStack, const T: usize> Minimq<N, T> {
+impl<
+        TcpStack: TcpClientStack,
+        Clock: embedded_time::Clock,
+        const MSG_SIZE: usize,
+        const MSG_COUNT: usize,
+    > Minimq<TcpStack, Clock, MSG_SIZE, MSG_COUNT>
+{
     /// Construct a new MQTT interface.
     ///
     /// # Args
@@ -404,11 +462,17 @@ impl<N: TcpClientStack, const T: usize> Minimq<N, T> {
     /// * `client_id` The client ID to use for communicating with the broker. If empty, rely on the
     ///   broker to automatically assign a client ID.
     /// * `network_stack` - The network stack to use for communication.
+    /// * `clock` - The clock to use for managing MQTT state timing.
     ///
     /// # Returns
     /// A `Minimq` object that can be used for publishing messages, subscribing to topics, and
     /// managing the MQTT state.
-    pub fn new(broker: IpAddr, client_id: &str, network_stack: N) -> Result<Self, Error<N::Error>> {
+    pub fn new(
+        broker: IpAddr,
+        client_id: &str,
+        network_stack: TcpStack,
+        clock: Clock,
+    ) -> Result<Self, Error<TcpStack::Error>> {
         let session_state = SessionState::new(
             broker,
             String::from_str(client_id).or(Err(Error::ProvidedClientIdTooLong))?,
@@ -416,8 +480,8 @@ impl<N: TcpClientStack, const T: usize> Minimq<N, T> {
 
         let minimq = Minimq {
             client: MqttClient {
-                network_stack,
-                socket: None,
+                network: InterfaceHolder::new(network_stack),
+                clock,
                 session_state,
                 connection_state: StateMachine::new(Context),
             },
@@ -432,21 +496,28 @@ impl<N: TcpClientStack, const T: usize> Minimq<N, T> {
     /// # Args
     /// * `f` - A closure to process any received messages. The closure should accept the client,
     /// topic, message, and list of proprties (in that order).
-    pub fn poll<F>(&mut self, mut f: F) -> Result<(), Error<N::Error>>
+    pub fn poll<F>(&mut self, mut f: F) -> Result<(), Error<TcpStack::Error>>
     where
-        for<'a> F: FnMut(&mut MqttClient<N, T>, &'a str, &[u8], &[Property<'a>]),
+        for<'a> F: FnMut(
+            &mut MqttClient<TcpStack, Clock, MSG_SIZE, MSG_COUNT>,
+            &'a str,
+            &[u8],
+            &[Property<'a>],
+        ),
     {
         self.client.process()?;
 
         // If the connection is no longer active, reset the packet reader state and return. There's
         // nothing more we can do.
-        if self.client.connection_state.state() != &States::Active {
+        if self.client.connection_state.state() != &States::Active
+            && self.client.connection_state.state() != &States::Establishing
+        {
             self.packet_reader.reset();
             return Ok(());
         }
 
         let mut buf: [u8; 1024] = [0; 1024];
-        let received = self.client.read(&mut buf)?;
+        let received = self.client.network.read(&mut buf)?;
         if received > 0 {
             debug!("Received {} bytes", received);
         }
@@ -468,16 +539,13 @@ impl<N: TcpClientStack, const T: usize> Minimq<N, T> {
 
             // Handle any received packets.
             while self.packet_reader.packet_available() {
-                let packet = ReceivedPacket::parse_message(&self.packet_reader)
-                    .map_err(|e| Error::Protocol(e))?;
+                let packet = ReceivedPacket::parse_message(&self.packet_reader)?;
 
                 info!("Received {:?}", packet);
 
                 let result = self.client.handle_packet(packet, &mut f);
 
-                self.packet_reader
-                    .pop_packet()
-                    .map_err(|e| Error::Protocol(e))?;
+                self.packet_reader.pop_packet()?;
 
                 // If there was an error, return it now. Note that we ensure the packet is removed
                 // from buffering after processing even in error conditions..
